""" PagerMaid module for adding captions to image. """

from os import remove
from pygments import highlight as syntax_highlight
from pygments.formatters import img
from pygments.lexers import guess_lexer
from pagermaid import command_help, log, log_chatid
from pagermaid.listener import listener, diagnostics
from pagermaid.utils import execute, obtain_source_file, upload_result_image


@listener(outgoing=True, command="convert")
@diagnostics
async def convert(context):
    """ Converts image to png. """
    try:
        reply_id, target_file_path = await obtain_source_file(context)
    except ValueError:
        return
    result = await execute("./utils/caption.sh \"" + target_file_path +
                           "\" result.png" + " \"" + str("") +
                           "\" " + "\"" + str("") + "\"")
    if not result:
        await context.edit("`Something wrong happened, please report this problem.`")
        try:
            remove("result.png")
            remove(target_file_path)
        except FileNotFoundError:
            pass
        return
    try:
        await context.client.send_file(
            context.chat_id,
            "result.png",
            reply_to=reply_id
        )
    except ValueError:
        await context.edit("`An error occurred during the conversion.`")
        remove(target_file_path)
        return
    await context.delete()
    remove(target_file_path)
    remove("result.png")


command_help.update({
    "convert": "Parameter: -convert <image>\
    \nUsage: Converts any image to png."
})


@listener(outgoing=True, command="caption")
@diagnostics
async def caption(context):
    """ Generates images with captions. """
    if context.pattern_match.group(1):
        if ',' in context.pattern_match.group(1):
            string_1, string_2 = context.pattern_match.group(1).split(',', 1)
        else:
            string_1 = context.pattern_match.group(1)
            string_2 = " "
    else:
        await context.edit("Invalid syntax.")
        return
    try:
        reply_id, target_file_path = await obtain_source_file(context)
    except ValueError:
        return
    result = await execute("./utils/caption.sh \"" + target_file_path +
                           "\" result.png" + " \"" + str(string_1) +
                           "\" " + "\"" + str(string_2) + "\"")
    try:
        await upload_result_image(context, result, target_file_path, reply_id)
    except ValueError:
        return
    message = string_1 + "` and `" + string_2
    if log:
        await context.client.send_message(
            log_chatid, "Captions `" + message + "` added to an image."
        )


command_help.update({
    "caption": "Parameter: -caption <text>,<text> <image>\
    \nUsage: Adds two lines of captions to an image."
})


@listener(outgoing=True, command="ocr")
@diagnostics
async def ocr(context):
    """ Extracts texts from images. """
    if context.fwd_from:
        return
    reply = await context.get_reply_message()
    await context.edit("`Processing image, please wait . . .`")
    if reply:
        target_file_path = await context.client.download_media(
            await context.get_reply_message()
        )
    else:
        target_file_path = await context.download_media()
    if target_file_path is None:
        await context.edit("`There are no attachment in target.`")
        return
    result = await execute(f"tesseract {target_file_path} stdout")
    if not result:
        await context.edit("`Something wrong happened, please report this problem.`")
        try:
            remove(target_file_path)
        except FileNotFoundError:
            pass
        return
    success = False
    if result == "/bin/sh: fbdump: command not found":
        await context.edit("A utility is missing.")
    else:
        result = await execute(f"tesseract {target_file_path} stdout", False)
        await context.edit(f"**Extracted text: **\n{result}")
        success = True
    remove(target_file_path)
    if not success:
        return
    if log:
        await context.client.send_message(
            log_chatid, "Performed OCR on an image."
        )


command_help.update({
    "ocr": "Parameter: -ocr <image>\
    \nUsage: Extracts texts from images."
})


@listener(outgoing=True, command="highlight")
@diagnostics
async def highlight(context):
    """ Generates syntax highlighted images. """
    if context.fwd_from:
        return
    reply = await context.get_reply_message()
    reply_id = None
    await context.edit("Rendering image, please wait . . .")
    if reply:
        reply_id = reply.id
        message = reply.text
    else:
        if context.pattern_match.group(1):
            message = context.pattern_match.group(1)
        else:
            await context.edit("`Unable to retrieve target message.`")
            return
    lexer = guess_lexer(message)
    formatter = img.JpgImageFormatter(style="colorful")
    result = syntax_highlight(message, lexer, formatter, outfile=None)
    await context.edit("Uploading image . . .")
    await context.client.send_file(
        context.chat_id,
        result,
        reply_to=reply_id
    )
    await context.delete()


command_help.update({
    "highlight": "Parameter: -highlight <image>\
    \nUsage: Generates syntax highlighted images."
})



""" This module handles world clock related utility. """

from datetime import datetime
from pytz import country_names
from pagermaid import command_help, config
from pagermaid.listener import listener, diagnostics
from pagermaid.utils import get_timezone


@listener(outgoing=True, command="time")
@diagnostics
async def time(context):
    """ For querying time. """
    country = context.pattern_match.group(1).title()
    time_form = "%I:%M %p"
    date_form = "%A %d/%m/%y"
    if not country:
        time_zone = await get_timezone(config['application_region'])
        await context.edit(
            f"**Time in {config['application_region']}**\n"
            f"`{datetime.now(time_zone).strftime(date_form)} "
            f"{datetime.now(time_zone).strftime(time_form)}`"
        )
        return

    time_zone = await get_timezone(country)
    if not time_zone:
        await context.edit("`Invalid parameter.`")
        return

    try:
        country_name = country_names[country]
    except KeyError:
        country_name = country

    await context.edit(f"**Time in {country_name}**\n"
                       f"`{datetime.now(time_zone).strftime(date_form)} "
                       f"{datetime.now(time_zone).strftime(time_form)}`")


command_help.update({
    "time": "Parameter: -time <region>\
    \nUsage: Displays time of specific region, reads from config file if parameter is empty."
})



""" QR Code related utilities. """

from os import remove
from pyqrcode import create
from PIL import Image
from pagermaid import command_help, log, log_chatid
from pagermaid.listener import listener, diagnostics

try:
    from pyzbar.pyzbar import decode
except ImportError:
    zBarFailure = True
    decode = None
    pass


@listener(outgoing=True, command="genqr")
@diagnostics
async def genqr(context):
    """ Generate QR codes. """
    downloaded_file_name = None
    if context.fwd_from:
        return
    input_string = context.pattern_match.group(1)
    message = None
    result = None
    if input_string:
        message = input_string
    elif context.reply_to_msg_id:
        reply = await context.get_reply_message()
        result = reply.id
        if reply.media:
            downloaded_file_name = await context.client.download_media(
                reply
            )
            with open(downloaded_file_name, "rb") as file:
                lines = file.readlines()
            message = ""
            for media in lines:
                try:
                    message += media.decode("UTF-8") + "\n"
                except UnicodeDecodeError:
                    await context.edit("`Unable to parse file as plaintext.`")
                    remove(downloaded_file_name)
                    return
            remove(downloaded_file_name)
        else:
            message = reply.message
    if message is None:
        await context.edit("`Invalid argument.`")
        return
    await context.edit("`Generating QR code.`")
    try:
        create(message, error='L', mode='binary').png('qr.webp', scale=6)
    except UnicodeEncodeError:
        await context.edit("`Invalid characters in target string.`")
        remove(downloaded_file_name)
        return
    await context.client.send_file(
        context.chat_id,
        "qr.webp",
        reply_to=result
    )
    remove("qr.webp")
    await context.delete()
    if log:
        await context.client.send_message(
            log_chatid, "Generated QR code for `" + message + "`."
        )
    await context.delete()


command_help.update({
    "genqr": "Parameter: -genqr <text>\
    \nUsage: Generates a QR code sticker."
})


@listener(outgoing=True, command="parseqr")
@diagnostics
async def parseqr(context):
    """ Parse QR code into plaintext. """
    if context.fwd_from:
        return
    if zBarFailure:
        await context.edit("`ZBar is not installed!`")
        return
    target_file_path = await context.client.download_media(
        await context.get_reply_message()
    )
    try:
        try:
            message = str(decode(Image.open(target_file_path))[0].data)[2:][:-1]
            await context.edit("Content: `" + message + "`.")
            success = True
        except IndexError:
            await context.edit("`Target is not a QR code.`")
            success = False
            message = None
        remove(target_file_path)
    except AttributeError:
        await context.edit("`Invalid argument.`")
        return
    if success:
        if log:
            await context.client.send_message(
                log_chatid, "Parsed QR code with content `" + message + "`."
            )
    else:
        if log:
            await context.client.send_message(
                log_chatid, "Attempted to parse non-encoded image as QR code."
            )


command_help.update({
    "parseqr": "Parameter: -parseqr\
    \nUsage: Parse the attached QR code into plaintext."
})



""" PagerMaid module for different ways to avoid users. """

from pagermaid import redis, redis_check, command_help, log, log_chatid
from pagermaid.listener import listener, diagnostics


@listener(outgoing=True, command="ghost")
@diagnostics
async def ghost(context):
    """ Toggles ghosting of a user. """
    if not redis_check():
        await context.edit("Redis is offline, cannot operate.")
        return
    if context.pattern_match.group(1) == 'true':
        redis.set("ghosted.chat_id." + str(context.chat_id), "true")
        await context.delete()
        if log:
            await context.client.send_message(
                log_chatid,
                "ChatID " + str(context.chat_id) + " added to ghosted chats.")
    elif context.pattern_match.group(1) == 'false':
        redis.delete("ghosted.chat_id." + str(context.chat_id))
        await context.delete()
        if log:
            await context.client.send_message(
                log_chatid,
                "ChatID " + str(context.chat_id) + " removed from ghosted chats.")
    elif context.pattern_match.group(1) == 'status':
        if redis.get("ghosted.chat_id." + str(context.chat_id)):
            await context.edit("Current chat is ghosted.")
        else:
            await context.edit("Current chat is not ghosted.")
    else:
        await context.edit("Invalid argument.")


command_help.update({
    "ghost": "Parameter: -ghost <true|false|status>\
    \nUsage: Toggles ghosting of chat, requires redis."
})


@listener(outgoing=True, command="deny")
@diagnostics
async def deny(context):
    """ Toggles denying of a user. """
    if not redis_check():
        await context.edit("Redis is offline, cannot operate.")
        return
    if context.pattern_match.group(1) == 'true':
        redis.set("denied.chat_id." + str(context.chat_id), "true")
        await context.delete()
        if log:
            await context.client.send_message(
                log_chatid,
                "ChatID " + str(context.chat_id) + " added to denied chats.")
    elif context.pattern_match.group(1) == 'false':
        redis.delete("denied.chat_id." + str(context.chat_id))
        await context.delete()
        if log:
            await context.client.send_message(
                log_chatid,
                "ChatID " + str(context.chat_id) + " removed from denied chats.")
    elif context.pattern_match.group(1) == 'status':
        if redis.get("silenced.chat_id." + str(context.chat_id)):
            await context.edit("Current chat is denied.")
        else:
            await context.edit("Current chat is not denied.")
    else:
        await context.edit("Invalid argument.")


command_help.update({
    "deny": "Parameter: -deny <true|false|status>\
    \nUsage: Toggles denying of chat, requires redis."
})


@listener(incoming=True, ignore_edited=True)
async def set_read_acknowledgement(context):
    """ Event handler to infinitely read ghosted messages. """
    if not redis_check():
        return
    if redis.get("ghosted.chat_id." + str(context.chat_id)):
        await context.client.send_read_acknowledge(context.chat_id)


@listener(incoming=True, ignore_edited=True)
async def message_removal(context):
    """ Event handler to infinitely delete denied messages. """
    if not redis_check():
        return
    if redis.get("denied.chat_id." + str(context.chat_id)):
        await context.delete()



""" System related utilities for PagerMaid to integrate into the system. """

from platform import node
from getpass import getuser
from os import remove
from os import geteuid
from pagermaid import command_help, log, log_chatid
from pagermaid.listener import listener, diagnostics
from pagermaid.utils import url_tracer, attach_log, execute


@listener(outgoing=True, command="evaluate")
@diagnostics
async def evaluate(context):
    """ Evaluate a python expression. """
    if context.is_channel and not context.is_group:
        await context.edit("`Evaluation is disabled in channels.`")
        return

    if context.pattern_match.group(1):
        expression = context.pattern_match.group(1)
    else:
        await context.edit("Invalid parameter.")
        return

    try:
        evaluation = str(eval(expression))
        if evaluation:
            if isinstance(evaluation, str):
                if len(evaluation) >= 4096:
                    file = open("output.log", "w+")
                    file.write(evaluation)
                    file.close()
                    await context.client.send_file(
                        context.chat_id,
                        "output.log",
                        reply_to=context.id,
                        caption="Output exceeded limit, attaching file.",
                    )
                    remove("output.log")
                    return
                await context.edit(
                    f">>> {expression}\n"
                    f"{evaluation}"
                )

    except Exception as err:
        await context.edit(
            f">>> {expression}\n"
            f"`{err}`"
        )

    if log:
        await context.client.send_message(
            log_chatid, f"Evaluated `{expression}` in the python interpreter."
        )


command_help.update({
    "evaluate": "Parameter: -evaluate <expression>>\
    \nUsage: Evaluate an expression in the python interpreter."
})


@listener(outgoing=True, command="sh")
@diagnostics
async def sh(context):
    """ For calling a binary from the shell. """
    user = getuser()
    command = context.pattern_match.group(1)
    uid = geteuid()
    hostname = node()
    if context.is_channel and not context.is_group:
        await context.edit("`Current configuration disables shell execution in channel.`")
        return

    if not command:
        await context.edit("`Invalid argument.`")
        return

    if uid == 0:
        await context.edit(
            f"`{user}`@{hostname} ~"
            f"\n> `#` {command}"
        )
    else:
        await context.edit(
            f"`{user}`@{hostname} ~"
            f"\n> `$` {command}"
        )

    result = await execute(command)

    if result:
        if len(result) > 4096:
            await attach_log(context, result)
            return

        if uid == 0:
            await context.edit(
                f"`{user}`@{hostname} ~"
                f"\n> `#` {command}"
                f"\n`{result}`"
            )
        else:
            await context.edit(
                f"`{user}`@{hostname} ~"
                f"\n> `$` {command}"
                f"\n`{result}`"
            )
    else:
        return
    if log:
        await context.client.send_message(
            log_chatid,
            "Command `" + command + "` executed.",
        )


command_help.update({
    "sh": "Parameter: -sh <command>\
    \nUsage: Executes a shell command."
})


@listener(outgoing=True, command="restart")
async def restart(context):
    """ To re-execute PagerMaid. """
    if not context.text[0].isalpha():
        await context.edit("Attempting system restart.")
        if log:
            await context.client.send_message(
                log_chatid,
                "PagerMaid restarted."
            )
        await context.client.disconnect()


command_help.update({
    "restart": "Parameter: -restart\
    \nUsage: Restarts PagerMaid."
})


@listener(outgoing=True, command="trace")
@diagnostics
async def trace(context):
    """ Trace URL redirects. """
    url = context.pattern_match.group(1)
    reply = await context.get_reply_message()
    if reply:
        url = reply.text
    if url:
        if url.startswith("https://") or url.startswith("http://"):
            pass
        else:
            url = "https://" + url
        await context.edit("`Tracing redirects . . .`")
        result = str("")
        for url in url_tracer(url):
            count = 0
            if result:
                result += " ↴\n" + url
            else:
                result = url
            if count == 128:
                result += "\n\nMore than 128 redirects, aborting!"
                break
        if result:
            if len(result) > 4096:
                await context.edit("`Output exceeded limit, attaching file.`")
                await attach_log(context, result)
                return
            await context.edit(
                "Redirects:\n"
                f"{result}"
            )
        else:
            await context.edit(
                "`Something wrong happened while making HTTP requests.`"
            )
    else:
        await context.edit("`Invalid argument.`")


command_help.update({
    "trace": "Parameter: -trace <url>\
    \nUsage: Traces redirect of a URL."
})


@listener(outgoing=True, command="contact")
@diagnostics
async def contact(context):
    """ To contact the creator of PagerMaid. """
    if not context.text[0].isalpha():
        await context.edit("`A conversation have been opened, click `[here](tg://user?id=503691334)` to enter.`",
                           parse_mode="markdown")
        message = "Hi, I would like to report something about PagerMaid."
        if context.pattern_match.group(1):
            message = context.pattern_match.group(1)
        await context.client.send_message(
            503691334,
            message
        )


command_help.update({
    "contact": "Parameter: -contact <message>\
    \nUsage: Contact the maintainer."
})


@listener(outgoing=True, command="exception")
@diagnostics
async def exception(context):
    """ Generates exception for debugging purposes. """
    raise ValueError('This is an exception generated by the system.')


command_help.update({
    "exception": "Parameter: -exception\
    \nUsage: Generates an exception for debugging."
})



""" The help module. """

from pagermaid import command_help
from pagermaid.listener import listener, diagnostics


@listener(outgoing=True, command="help")
@diagnostics
async def help(context):
    """ The help command,"""
    args = context.pattern_match.group(1)
    if args:
        if args in command_help:
            await context.edit(str(command_help[args]))
        else:
            await context.edit("`Invalid argument, please check module list.`")
    else:
        result = "**Commands loaded from current modules: \n**"
        for i in sorted(command_help, reverse=False):
            result += "`" + str(i)
            result += "`, "
        await context.edit(result[:-2] + "\n**Do \"-help <command>\" to view help for a specific command.**")


command_help.update({
    "help": "Parameter: -help <command>\
    \nUsage: Shows a list of commands or help string of a single command."
})



""" PagerMaid module to handle sticker collection. """

from urllib import request
from io import BytesIO
from telethon.tl.types import DocumentAttributeFilename, MessageMediaPhoto
from pagermaid import bot, command_help
from pagermaid.listener import listener, diagnostics
from pagermaid.utils import add_sticker, resize_image, upload_sticker


@listener(outgoing=True, command="sticker")
@diagnostics
async def sticker(context):
    """ Fetches images/stickers and add them to your pack. """
    user = await bot.get_me()
    if not user.username:
        user.username = user.first_name
    message = await context.get_reply_message()
    custom_emoji = False
    animated = False
    emoji = ""
    await context.edit("Collecting sticker . . .")
    if message and message.media:
        if isinstance(message.media, MessageMediaPhoto):
            photo = BytesIO()
            photo = await bot.download_media(message.photo, photo)
        elif "image" in message.media.document.mime_type.split('/'):
            photo = BytesIO()
            await context.edit("Downloading image . . .")
            await bot.download_file(message.media.document, photo)
            if (DocumentAttributeFilename(file_name='sticker.webp') in
                    message.media.document.attributes):
                emoji = message.media.document.attributes[1].alt
                custom_emoji = True
        elif (DocumentAttributeFilename(file_name='AnimatedSticker.tgs') in
              message.media.document.attributes):
            emoji = message.media.document.attributes[0].alt
            custom_emoji = True
            animated = True
            photo = 1
        else:
            await context.edit("`This file type is not supported.`")
            return
    else:
        await context.edit("`Please reply to a message with an image/sticker.`")
        return

    if photo:
        split_strings = context.text.split()
        if not custom_emoji:
            emoji = "👀"
        pack = 1
        if len(split_strings) == 3:
            pack = split_strings[2]
            emoji = split_strings[1]
        elif len(split_strings) == 2:
            if split_strings[1].isnumeric():
                pack = int(split_strings[1])
            else:
                emoji = split_strings[1]

        pack_name = f"pack_{user.id}_{user.username}_{pack}"
        pack_title = f"@{user.username}'s collection ({pack})"
        command = '/newpack'
        file = BytesIO()

        if not animated:
            await context.edit("Resizing image . . .")
            image = await resize_image(photo)
            file.name = "sticker.png"
            image.save(file, "PNG")
        else:
            pack_name += "_animated"
            pack_title += " (animated)"
            command = '/newanimated'

        response = request.urlopen(
            request.Request(f'http://t.me/addstickers/{pack_name}'))
        http_response = response.read().decode("utf8").split('\n')

        if "  A <strong>Telegram</strong> user has created the <strong>Sticker&nbsp;Set</strong>." not in \
                http_response:
            async with bot.conversation('Stickers') as conversation:
                await conversation.send_message('/addsticker')
                await conversation.get_response()
                await bot.send_read_acknowledge(conversation.chat_id)
                await conversation.send_message(pack_name)
                chat_response = await conversation.get_response()
                while chat_response.text == "Whoa! That's probably enough stickers for one pack, give it a break. \
A pack can't have more than 120 stickers at the moment.":
                    pack += 1
                    pack_name = f"a{user.id}_by_{user.username}_{pack}"
                    pack_title = f"@{user.username}'s collection ({pack})"
                    await context.edit("Switching to pack " + str(pack) +
                                       " since previous pack is full . . .")
                    await conversation.send_message(pack_name)
                    chat_response = await conversation.get_response()
                    if chat_response.text == "Invalid pack selected.":
                        await add_sticker(conversation, command, pack_title, pack_name, bot, animated, message,
                                          context, file, emoji)
                        await context.edit(
                            f"Sticker has been added to [this](t.me/addstickers/{pack_name}) alternative pack.",
                            parse_mode='md')
                        return
                await upload_sticker(animated, bot, message, context, file, conversation)
                await conversation.get_response()
                await conversation.send_message(emoji)
                await bot.send_read_acknowledge(conversation.chat_id)
                await conversation.get_response()
                await conversation.send_message('/done')
                await conversation.get_response()
                await bot.send_read_acknowledge(conversation.chat_id)
        else:
            await context.edit("Pack does not exist, creating . . .")
            async with bot.conversation('Stickers') as conversation:
                await add_sticker(conversation, command, pack_title, pack_name, bot, animated, message,
                                  context, file, emoji)

        await context.edit(
            f"Sticker has been added to [this](t.me/addstickers/{pack_name}) pack.",
            parse_mode='md')


command_help.update({
    "sticker": "Parameter: -sticker <emoji>\
    \nUsage: Collects image/sticker as sticker, specify emoji to set custom emoji."
})



""" Fun related chat utilities. """

from asyncio import sleep
from random import choice
from telethon.errors.rpcerrorlist import MessageNotModifiedError
from cowpy import cow
from pagermaid import command_help, bot
from pagermaid.listener import listener, diagnostics
from pagermaid.utils import mocker, corrupt, owoifier, execute, random_gen


@listener(command="animate")
@diagnostics
async def animate(context):
    """ Command for animated texts. """
    text = await context.get_reply_message()
    message = context.pattern_match.group(1)
    if message:
        pass
    elif text:
        message = text.text
    else:
        await context.edit("`Invalid argument.`")
        return
    latency = 0.03
    cursor = "█"
    old_text = ''
    await context.edit(cursor)
    await sleep(latency)
    for character in message:
        old_text = old_text + "" + character
        typing_text = old_text + "" + cursor
        await context.edit(typing_text)
        await sleep(latency)
        await context.edit(old_text)
        await sleep(latency)


command_help.update({
    "animate": "Parameter: -animate <text>\
    \nUsage: Animated text via edit."
})


@listener(outgoing=True, command="mock")
@diagnostics
async def mock(context):
    """ Mock people with weird caps. """
    reply = await context.get_reply_message()
    message = context.pattern_match.group(1)
    if message:
        pass
    elif reply:
        message = reply.text
    else:
        await context.edit("`Invalid arguments.`")
        return

    reply_text = mocker(message)
    if reply:
        if reply.sender.is_self:
            try:
                await reply.edit(reply_text)
            except MessageNotModifiedError:
                await context.edit("`This message is already mocked in the same way.`")
                return
            await context.delete()
    else:
        await context.edit(reply_text)


command_help.update({
    "mock": "Parameter: -mock <text>\
    \nUsage: Mock a string via weird caps."
})


@listener(outgoing=True, command="widen")
@diagnostics
async def widen(context):
    """ Make texts weirdly wide. """
    reply = await context.get_reply_message()
    message = context.pattern_match.group(1)
    if message:
        pass
    elif reply:
        message = reply.text
    else:
        await context.edit("`Invalid argument.`")
        return

    reply_text = str(message).translate(dict((i, i + 0xFEE0) for i in range(0x21, 0x7F)))
    if reply:
        if reply.sender.is_self:
            try:
                await reply.edit(reply_text)
            except MessageNotModifiedError:
                await context.edit("`This message is already widened.`")
                return
            await context.delete()
    else:
        await context.edit(reply_text)


command_help.update({
    "widen": "Parameter: -widen <text>\
    \nUsage: Widen every char in a string in a weird way."
})


@listener(outgoing=True, command="fox")
@diagnostics
async def fox(context):
    """ Make a fox scratch your message. """
    reply = await context.get_reply_message()
    message = context.pattern_match.group(1)
    if message:
        pass
    elif reply:
        message = reply.text
    else:
        await context.edit(
            "`Invalid arguments.`"
        )
        return

    input_text = " ".join(message).lower()
    reply_text = corrupt(input_text)
    if reply:
        if reply.sender.is_self:
            try:
                await reply.edit(reply_text)
            except MessageNotModifiedError:
                await context.edit("`Message have already been scratched.`")
            await context.delete()
        else:
            await context.edit(reply_text)


command_help.update({
    "fox": "Parameter: -fox <text>\
    \nUsage: Make a fox corrupt your text."
})


@listener(outgoing=True, command="owo")
@diagnostics
async def owo(context):
    """ Makes messages become owo. """
    if context.pattern_match.group(1):
        text = context.pattern_match.group(1)
        await context.edit(owoifier(text))
        return
    elif context.reply_to_msg_id:
        reply_msg = await context.get_reply_message()
        if reply_msg.sender.is_self:
            try:
                result = owoifier(reply_msg.text)
            except BaseException:
                await context.delete()
                return
            await reply_msg.edit(result)
            await context.delete()
        else:
            try:
                result = owoifier(reply_msg.text)
            except BaseException:
                await context.delete()
                return
            await context.edit(result)
        return
    else:
        await context.edit("`Unable to get the target message.`")
        return


command_help.update({
    "owo": "Parameter: -owo <text>\
    \nUsage: Converts messages to OwO."
})


@listener(outgoing=True, command="ship")
@diagnostics
async def ship(context):
    """ Ship randomly generated members. """
    await context.edit("`Generating couple . . .`")
    users = []
    async for user in context.client.iter_participants(context.chat_id):
        users.append(user)
    target_1 = choice(users)
    target_2 = choice(users)
    if context.pattern_match.group(1):
        if ' ' in context.pattern_match.group(1):
            string_1, string_2 = context.pattern_match.group(1).split(' ', 1)
            user = string_1
            if user.isnumeric():
                user = int(user)
            try:
                target_1 = await context.client.get_entity(user)
            except (TypeError, ValueError) as err:
                await context.edit(str(err))
                return None
            user = string_2
            if user.isnumeric():
                user = int(user)
            try:
                target_2 = await context.client.get_entity(user)
            except (TypeError, ValueError) as err:
                await context.edit(str(err))
                return None
        else:
            string_1 = context.pattern_match.group(1)
            user = string_1
            if user.isnumeric():
                user = int(user)
            try:
                target_1 = await context.client.get_entity(user)
            except (TypeError, ValueError) as err:
                await context.edit(str(err))
                return None
    if len(users) == 1:
        target_1 = users[0]
        target_2 = await bot.get_me()
    await context.edit("**Generated couple**\n"
                       + f"[{target_1.first_name}](tg://user?id={target_1.id})" + " + "
                       + f"[{target_2.first_name}](tg://user?id={target_2.id})" + " = " + "❤️")


command_help.update({
    "ship": "Parameter: -ship <user> <user>\
    \nUsage: Ships random person, supports defining person to ship."
})


@listener(outgoing=True, command="aaa")
@diagnostics
async def aaa(context):
    """ Generates screams. """
    await random_gen(context, "Aa")


command_help.update({
    "aaa": "Parameter: -aaa <integer>\
    \nUsage: Generates screams."
})


@listener(outgoing=True, command="asciiart")
@diagnostics
async def asciiart(context):
    """ Generates ASCII art for provided strings. """
    if not context.pattern_match.group(1):
        await context.edit("`Invalid arguments.`")
        return
    output = await execute(f"figlet -f ./utils/graffiti.flf '{context.pattern_match.group(1)}'")
    await context.edit(f"`{output}`")


command_help.update({
    "asciiart": "Parameter: -asciiart <text>\
     \nUsage: Generates ASCII art for target text."
})


@listener(outgoing=True, command="tuxsay")
@diagnostics
async def tuxsay(context):
    """ Generates ASCII art for Tux saying stuff. """
    if not context.pattern_match.group(1):
        await context.edit("`Invalid arguments.`")
        return
    await context.edit(f"```{cow.Tux().milk(context.pattern_match.group(1))} \n ```")


command_help.update({
    "asciiart": "Parameter: -asciiart <text>\
     \nUsage: Generates ASCII art for target text."
})


@listener(outgoing=True, pattern="^OwO$")
async def wink(context):
    """ OwO """
    await context.edit("OwU")
    await sleep(.3)
    await context.edit("OwO")


command_help.update({
    "wink": "Parameter: OwO\
     \nUsage: Send OwO in chat."
})


@listener(outgoing=True, command="coin")
async def coin(context):
    await context.edit("Throwing coin . . .")
    await sleep(.5)
    outcomes = ['A'] * 5 + ['B'] * 5 + ['C'] * 1
    result = choice(outcomes)
    count = 0
    while count <= 3:
        await context.edit("`.` . .")
        await sleep(.3)
        await context.edit(". `.` .")
        await sleep(.3)
        await context.edit(". . `.`")
        await sleep(.3)
        count += 1
    if result == "C":
        await context.edit("WAAAHHH I LOST THE COIN")
    elif result == "B":
        await context.edit("Tails!")
    elif result == "A":
        await context.edit("Heads!")

command_help.update({
    "coin": "Parameter: -coin\
     \nUsage: Throws a coin."
})



""" PagerMaid features that uses external HTTP APIs other than Telegram. """

from googletrans import Translator, LANGUAGES
from os import remove
from gtts import gTTS
from re import compile as regex_compile
from pagermaid.utils import fetch_youtube_audio
from pagermaid import command_help, log, log_chatid
from pagermaid.listener import listener, diagnostics, config
from pagermaid.utils import clear_emojis, attach_log, GoogleSearch


@listener(outgoing=True, command="translate")
@diagnostics
async def translate(context):
    """ PagerMaid universal translator. """
    translator = Translator()
    text = await context.get_reply_message()
    message = context.pattern_match.group(1)
    lang = config['lang']
    if message:
        pass
    elif text:
        message = text.text
    else:
        await context.edit("`Invalid parameter.`")
        return

    try:
        await context.edit("`Generating translation . . .`")
        result = translator.translate(clear_emojis(message), dest=lang)
    except ValueError:
        await context.edit("`Language not found, please correct the error in the config file.`")
        return

    source_lang = LANGUAGES[f'{result.src.lower()}']
    trans_lang = LANGUAGES[f'{result.dest.lower()}']
    result = f"**Translated** from {source_lang.title()}:\n{result.text}"

    if len(result) > 4096:
        await context.edit("`Output exceeded limit, attaching file.`")
        await attach_log(context, result)
        return
    await context.edit(result)
    if log:
        result = f"Translated `{message}` from {source_lang} to {trans_lang}."
        if len(result) > 4096:
            await context.edit("`Output exceeded limit, attaching file.`")
            await attach_log(context, result)
            return
        await context.client.send_message(
            log_chatid,
            result,
        )


command_help.update({
    "translate": "Parameter: -translate <text>\
    \nUsage: Translate the target message into English."
})


@listener(outgoing=True, command="tts")
@diagnostics
async def tts(context):
    """ Send TTS stuff as voice message. """
    text = await context.get_reply_message()
    message = context.pattern_match.group(1)
    lang = config['lang']
    if message:
        pass
    elif text:
        message = text.text
    else:
        await context.edit("`Invalid argument.`")
        return

    try:
        await context.edit("`Generating vocals . . .`")
        gTTS(message, lang)
    except AssertionError:
        await context.edit("`Invalid argument.`")
        return
    except ValueError:
        await context.edit('`Language not found, please correct the error in the config file.`')
        return
    except RuntimeError:
        await context.edit('`Error loading array of languages.`')
        return
    gtts = gTTS(message, lang)
    gtts.save("vocals.mp3")
    with open("vocals.mp3", "rb") as audio:
        line_list = list(audio)
        line_count = len(line_list)
    if line_count == 1:
        gtts = gTTS(message, lang)
        gtts.save("vocals.mp3")
    with open("vocals.mp3", "r"):
        await context.client.send_file(context.chat_id, "vocals.mp3", voice_note=True)
        remove("vocals.mp3")
        if log:
            await context.client.send_message(
                log_chatid, "Generated tts for `" + message + "`."
            )
        await context.delete()


command_help.update({
    "tts": "Parameter: -tts <text>\
    \nUsage: Generates a voice message."
})


@listener(outgoing=True, command="google")
@diagnostics
async def google(context):
    """ Searches Google for a string. """
    query = context.pattern_match.group(1)
    await context.edit("Pulling results . . .")
    search_results = GoogleSearch().search(query=query)
    results = ""
    count = 0
    for result in search_results.results:
        if count == int(config['result_length']):
            break
        count += 1
        title = result.title
        link = result.url
        desc = result.text
        results += f"\n[{title}]({link}) \n`{desc}`\n"
    await context.edit(f"**Google** |`{query}`| 🎙 🔍 \n"
                       f"{results}",
                       link_preview=False)
    if log:
        await context.client.send_message(
            log_chatid,
            "Queried `" + query + "` on Google Search.",
        )


command_help.update({
    "google": "Parameter: -google <text>\
    \nUsage: Searches Google for a string."
})


@listener(outgoing=True, command="fetchaudio")
@diagnostics
async def fetchaudio(context):
    """ Fetches audio from provided URL. """
    url = context.pattern_match.group(1)
    reply = await context.get_reply_message()
    reply_id = None
    await context.edit("Fetching audio . . .")
    if reply:
        reply_id = reply.id
    if url is None:
        await context.edit("Invalid arguments.")
        return
    youtube_pattern = regex_compile(r"^(http(s)?://)?((w){3}.)?youtu(be|.be)?(\.com)?/.+")
    if youtube_pattern.match(url):
        await fetch_youtube_audio(context, url, reply_id)
        if log:
            await context.client.send_message(
                log_chatid, "Fetched audio from `" + url + "`."
            )


command_help.update({
    "fetchaudio": "Parameter: -fetchaudio <url>\
    \nUsage: Fetches audio from multiple platforms."
})



""" Module that show system info of the hardware the bot is running on. """

from os import remove
from datetime import datetime
from speedtest import Speedtest
from telethon import functions
from platform import python_version, uname
from shutil import which
from telethon import version as telethon_version
from sys import platform
from pagermaid import command_help, redis_check
from pagermaid.listener import listener, diagnostics
from pagermaid.utils import unit_convert, execute, make_top_cloud

hostname = uname().node
kernel = uname().release


@listener(outgoing=True, command="sysinfo")
@diagnostics
async def sysinfo(context):
    """ Fetches system info using neofetch. """
    result = await execute("neofetch --config none --stdout")
    await context.edit("`" + result + "`")


command_help.update({
    "sysinfo": "Parameter: -sysinfo\
    \nUsage: Retrieve system information via neofetch."
})


@listener(outgoing=True, command="fortune")
@diagnostics
async def fortune(context):
    """ Reads a fortune cookie. """
    result = await execute("fortune")
    if result == "/bin/sh: fortune: command not found":
        await context.edit("`No fortune cookies on this system.`")
        return
    await context.edit(result)


command_help.update({
    "fortune": "Parameter: -fortune\
    \nUsage: Reads a fortune cookie message."
})


@listener(outgoing=True, command="tty")
@diagnostics
async def tty(context):
    """ Screenshots a TTY and prints it. """
    await context.edit("`Taking screenshot of framebuffer . . .`")
    message_id_to_reply = context.message.reply_to_msg_id
    if not message_id_to_reply:
        message_id_to_reply = None
    result = await execute("fbdump | magick - image.png")
    if result == "/bin/sh: fbdump: command not found":
        await context.edit("`fbdump does not exist on this system.`")
        remove("image.png")
        return
    if result == "/bin/sh: convert: command not found":
        await context.edit("`ImageMagick does not exist on this system.`")
        remove("image.png")
        return
    if result == "Failed to open /dev/fb0: Permission denied":
        await context.edit("`User not in video group.`")
        remove("image.png")
        return
    try:
        await context.client.send_file(
            context.chat_id,
            "image.png",
            caption="Screenshot of currently attached tty.",
            link_preview=False,
            force_document=False,
            reply_to=message_id_to_reply
        )
    except ValueError:
        await context.edit("`File is not generated due to unexpected error.`")
        return
    await context.delete()
    remove("image.png")


command_help.update({
    "tty": "Parameter: -tty\
    \nUsage: Takes screenshot of a TTY."
})


@listener(outgoing=True, command="version")
@diagnostics
async def version(context):
    """ Command to query the version of PagerMaid. """
    if which("git") is not None:
        version_result = await execute("git describe --all --long")
        revision_result = await execute("git rev-list --all --count")

        await context.edit(
            "`PagerMaid Version: "
            f"{version_result}"
            "` \n"
            "`Git Revision: "
            f"{revision_result}"
            "`"
        )
    else:
        await context.edit(
            "Git is malfunctioning."
        )


command_help.update({
    "version": "Parameter: -version\
    \nUsage: Outputs the version and git revision."
})


@listener(outgoing=True, command="status")
@diagnostics
async def status(context):
    if not redis_check():
        db = "Redis is malfunctioning."
    else:
        db = "Connected to Redis."
    await context.edit(
        "`"
        "PagerMaid is online. \n\n"
        f"Hostname: {hostname} \n"
        f"Database Status: {db} \n"
        f"Host Platform: {platform} \n"
        f"Kernel Version: {kernel} \n"
        f"Python Version: {python_version()} \n"
        f"Library version: {telethon_version.__version__}"
        "`"
    )


command_help.update({
    "status": "Parameter: -status\
    \nUsage: Output the status of PagerMaid"
})


@listener(outgoing=True, command="speedtest")
@diagnostics
async def speedtest(context):
    """ Tests internet speed using speedtest. """
    result = None
    await context.edit("`Executing test scripts . . .`")
    test = Speedtest()

    test.get_best_server()
    test.download()
    test.upload()
    test.results.share()
    result = test.results.dict()

    await context.edit("Timestamp "
                       f"`{result['timestamp']}` \n\n"
                       "Upload "
                       f"`{unit_convert(result['upload'])}` \n"
                       "Download "
                       f"`{unit_convert(result['download'])}` \n"
                       "Latency "
                       f"`{result['ping']}` \n"
                       "ISP "
                       f"`{result['client']['isp']}`")


command_help.update({
    "speedtest": "Parameter: -speedtest\
    \nUsage: Execute the speedtest script and outputs your internet speed."
})


@listener(outgoing=True, command="connection")
@diagnostics
async def connection(context):
    """ Shows connection info. """
    datacenter = await context.client(functions.help.GetNearestDcRequest())
    await context.edit(
        f"Region `{datacenter.country}` \n"
        f"Connected Datacenter `{datacenter.this_dc}` \n"
        f"Nearest Datacenter `{datacenter.nearest_dc}`"
    )


command_help.update({
    "connection": "Parameter: -connection\
    \nUsage: Shows your connection info."
})


@listener(outgoing=True, command="ping")
@diagnostics
async def ping(context):
    """ Calculates the latency of the bot host. """
    start = datetime.now()
    await context.edit("`Pong!`")
    end = datetime.now()
    duration = (end - start).microseconds / 1000
    await context.edit("`Pong!|%sms`" % duration)


command_help.update({
    "ping": "Parameter: -ping\
    \nUsage: Outputs your latency to telegram."
})


@listener(outgoing=True, command="topcloud")
@diagnostics
async def topcloud(context):
    """ Generates a word cloud of resource-hungry processes. """
    if context.fwd_from:
        return
    await make_top_cloud(context)
    await context.edit("Uploading image . . .")
    await context.client.send_file(
        context.chat_id,
        "cloud.png",
        reply_to=None,
        caption="Cloud of running processes."
    )
    remove("cloud.png")
    await context.delete()


command_help.update({
    "topcloud": "Parameter: -topcloud <image>\
    \nUsage: Generates a word cloud of resource-hungry processes."
})



""" PagerMaid launch sequence. """

from importlib import import_module
from telethon.errors.rpcerrorlist import PhoneNumberInvalidError
from pagermaid import logs, bot
from pagermaid.modules import all_modules

invalid_phone = '\nInvalid phone number entered.' \
                '\nPlease make sure you specified' \
                '\nyour country code in the string.'

try:
    bot.start()
except PhoneNumberInvalidError:
    print(invalid_phone)
    exit(1)

for module_name in all_modules:
    try:
        imported_module = import_module("pagermaid.modules." + module_name)
    except BaseException:
        logs.info(f"Unable to load module {module_name}.")

logs.info("PagerMaid have started, The prefix is -, type -help for help message.")

bot.run_until_disconnected()



""" Libraries for python modules. """

from os import remove, popen
from emoji import get_emoji_regexp
from requests import head
from requests.exceptions import MissingSchema, InvalidURL, ConnectionError
from wordcloud import WordCloud
from urllib import request, parse
from PIL import Image
from math import floor, ceil
from bs4 import BeautifulSoup
from random import random, randint, randrange, seed, choice
from time import sleep
from threading import Thread
from json import load as load_json
from re import sub, search, IGNORECASE
from pytz import country_names
from telethon.tl.functions.users import GetFullUserRequest
from telethon.tl.types import MessageEntityMentionName
from pytz import timezone, country_timezones
from asyncio import create_subprocess_shell
from asyncio.subprocess import PIPE
from pathlib import Path
from collections import deque
from youtube_dl import YoutubeDL
from pagermaid import config


async def upload_result_image(context, result, target_file_path, reply_id):
    if not result:
        await context.edit("`Something wrong happened, please report this problem.`")
        try:
            remove("result.png")
            remove(target_file_path)
        except FileNotFoundError:
            pass
        raise ValueError
    try:
        await context.client.send_file(
            context.chat_id,
            "result.png",
            reply_to=reply_id
        )
    except ValueError:
        await context.edit("`An error occurred during the conversion.`")
        remove(target_file_path)
        raise ValueError
    await context.delete()
    remove(target_file_path)
    remove("result.png")


async def obtain_source_file(context):
    if context.fwd_from:
        return
    reply = await context.get_reply_message()
    reply_id = None
    await context.edit("Rendering image, please wait . . .")
    if reply:
        reply_id = reply.id
        target_file_path = await context.client.download_media(
            await context.get_reply_message()
        )
    else:
        target_file_path = await context.download_media()
    if target_file_path is None:
        await context.edit("`There are no attachment in target.`")
        raise ValueError
    return reply_id, target_file_path


async def fetch_youtube_audio(context, url, reply_id):
    youtube_dl_options = {
        'format': 'bestaudio/best',
        'outtmpl': "output.%(ext)s",
        'postprocessors': [{
            'key': 'FFmpegExtractAudio',
            'preferredcodec': 'mp3',
            'preferredquality': '192',
        }],
    }
    YoutubeDL(youtube_dl_options).download([url])
    try:
        await context.client.send_file(
            context.chat_id,
            "output.mp3",
            reply_to=reply_id
        )
    except ValueError:
        await context.edit("`An error occurred during the download.`")
        return
    await context.delete()
    remove("output.mp3")


async def make_top_cloud(context):
    await context.edit("Generating image . . .")
    command_list = []
    if not Path('/usr/bin/top').is_symlink():
        output = str(await execute("top -b -n 1")).split("\n")[7:]
    else:
        output = str(await execute("top -b -n 1")).split("\n")[4:]
    for line in output[:-1]:
        line = sub(r'\s+', ' ', line).strip()
        fields = line.split(" ")
        try:
            if fields[11].count("/") > 0:
                command = fields[11].split("/")[0]
            else:
                command = fields[11]

            cpu = float(fields[8].replace(",", "."))
            mem = float(fields[9].replace(",", "."))

            if command != "top":
                command_list.append((command, cpu, mem))
        except BaseException:
            pass
    command_dict = {}
    for command, cpu, mem in command_list:
        if command in command_dict:
            command_dict[command][0] += cpu
            command_dict[command][1] += mem
        else:
            command_dict[command] = [cpu + 1, mem + 1]

    resource_dict = {}

    for command, [cpu, mem] in command_dict.items():
        resource_dict[command] = (cpu ** 2 + mem ** 2) ** 0.5

    width, height = None, None
    try:
        width, height = ((popen("xrandr | grep '*'").read()).split()[0]).split("x")
        width = int(width)
        height = int(height)
    except BaseException:
        pass
    if not width or not height:
        width = int(config['width'])
        height = int(config['height'])
    background = config['background']
    margin = int(config['margin'])

    cloud = WordCloud(
        background_color=background,
        width=width - 2 * int(margin),
        height=height - 2 * int(margin)
    ).generate_from_frequencies(resource_dict)

    cloud.to_file("cloud.png")


async def upload_sticker(animated, bot, message, context, file, conversation):
    if animated:
        await bot.forward_messages(
            'Stickers', [message.id], context.chat_id)
    else:
        file.seek(0)
        await context.edit("Uploading image . . .")
        await conversation.send_file(file, force_document=True)


async def add_sticker(conversation, command, pack_title, pack_name, bot, animated, message, context, file, emoji):
    await conversation.send_message(command)
    await conversation.get_response()
    await bot.send_read_acknowledge(conversation.chat_id)
    await conversation.send_message(pack_title)
    await conversation.get_response()
    await bot.send_read_acknowledge(conversation.chat_id)
    await upload_sticker(animated, bot, message, context, file, conversation)
    await conversation.get_response()
    await conversation.send_message(emoji)
    await bot.send_read_acknowledge(conversation.chat_id)
    await conversation.get_response()
    await conversation.send_message("/publish")
    if animated:
        await conversation.get_response()
        await conversation.send_message(f"<{pack_title}>")
    await conversation.get_response()
    await bot.send_read_acknowledge(conversation.chat_id)
    await conversation.send_message("/skip")
    await bot.send_read_acknowledge(conversation.chat_id)
    await conversation.get_response()
    await conversation.send_message(pack_name)
    await bot.send_read_acknowledge(conversation.chat_id)
    await conversation.get_response()
    await bot.send_read_acknowledge(conversation.chat_id)


async def execute(command, pass_error=True):
    executor = await create_subprocess_shell(
        command,
        stdout=PIPE,
        stderr=PIPE
    )

    stdout, stderr = await executor.communicate()
    if pass_error:
        result = str(stdout.decode().strip()) \
                 + str(stderr.decode().strip())
    else:
        result = str(stdout.decode().strip())
    return result


async def random_gen(context, selection):
    try:
        length = context.pattern_match.group(1)
        if not length:
            length = 64
        result = await execute(f"head -c 65536 /dev/urandom | tr -dc {selection} | head -c {length} ; echo \'\'")
        await context.edit(result)
    except FileNotFoundError:
        await context.edit("`Kat didn't bring /usr/bin/head.`")


async def changelog_gen(repo, diff):
    result = ''
    d_form = "%d/%m/%y"
    for c in repo.iter_commits(diff):
        result += f'•[{c.committed_datetime.strftime(d_form)}]: {c.summary} <{c.author}>\n'
    return result


async def branch_check(branch):
    official = ['master', 'staging']
    for k in official:
        if k == branch:
            return 1
    return


async def send_prune_notify(context, count):
    return await context.client.send_message(
        context.chat_id,
        "Deleted "
        + str(count)
        + " messages."
    )


async def attach_log(context, result):
    """ Method to attach logs for messages that are too long. """
    file = open("output.log", "w+")
    file.write(result)
    file.close()
    await context.client.send_file(
        context.chat_id,
        "output.log",
        reply_to=context.id,
    )
    remove("output.log")


async def get_timezone(target):
    """ Returns timezone of the parameter in command. """
    if "(Uk)" in target:
        target = target.replace("Uk", "UK")
    if "(Us)" in target:
        target = target.replace("Us", "US")
    if " Of " in target:
        target = target.replace(" Of ", " of ")
    if "(Western)" in target:
        target = target.replace("(Western)", "(western)")
    if "Minor Outlying Islands" in target:
        target = target.replace("Minor Outlying Islands", "minor outlying islands")
    if "Nl" in target:
        target = target.replace("Nl", "NL")

    for country_code in country_names:
        if target == country_names[country_code]:
            return timezone(country_timezones[country_code][0])
    try:
        if country_names[target]:
            return timezone(country_timezones[target][0])
    except KeyError:
        return


async def resize_image(photo):
    """ Photo resize to match sticker standards. """
    image = Image.open(photo)
    maxsize = (512, 512)
    if (image.width and image.height) < 512:
        size1 = image.width
        size2 = image.height
        if image.width > image.height:
            scale = 512 / size1
            size1new = 512
            size2new = size2 * scale
        else:
            scale = 512 / size2
            size1new = size1 * scale
            size2new = 512
        size1new = floor(size1new)
        size2new = floor(size2new)
        sizenew = (size1new, size2new)
        image = image.resize(sizenew)
    else:
        image.thumbnail(maxsize)

    return image


async def fetch_user(target):
    """ Fetch information of the target user. """
    if target.reply_to_msg_id:
        previous_message = await target.get_reply_message()
        replied_user = await target.client(GetFullUserRequest(previous_message.from_id))
    else:
        user = target.pattern_match.group(1)

        if user.isnumeric():
            user = int(user)

        if not user:
            self_user = await target.client.get_me()
            user = self_user.id

        if target.message.entities is not None:
            probable_user_mention_entity = target.message.entities[0]

            if isinstance(probable_user_mention_entity, MessageEntityMentionName):
                user_id = probable_user_mention_entity.user_id
                replied_user = await target.client(GetFullUserRequest(user_id))
                return replied_user
        try:
            user_object = await target.client.get_entity(user)
            replied_user = await target.client(GetFullUserRequest(user_object.id))
        except (TypeError, ValueError) as err:
            await target.edit(str(err))
            return None

    return replied_user


async def generate_strings(replied_user):
    """ Generates the needed strings for a user. """
    user_id = replied_user.user.id
    first_name = replied_user.user.first_name
    last_name = replied_user.user.last_name
    common_chat = replied_user.common_chats_count
    user_name = replied_user.user.username
    user_bio = replied_user.about
    is_bot = replied_user.user.bot
    restricted = replied_user.user.restricted
    verified = replied_user.user.verified
    first_name = first_name.replace("\u2060", "") if first_name else (
        "This user does not have a first name.")
    last_name = last_name.replace("\u2060", "") if last_name else (
        "This user does not have a last name.")
    user_name = "@{}".format(user_name) if user_name else (
        "This user does not have a username.")
    user_bio = "This user has no bio." if not user_bio else user_bio

    caption = "**Profile:** \n"
    caption += f"First Name: {first_name} \n"
    caption += f"Last Name: {last_name} \n"
    caption += f"Username: {user_name} \n"
    caption += f"Bot: {is_bot} \n"
    caption += f"Restricted: {restricted} \n"
    caption += f"Verified: {verified} \n"
    caption += f"ID: `{user_id}` \n"
    caption += f"Bio: `{user_bio}` \n"
    caption += f"Common Chats: {common_chat} \n"
    caption += f"Permanent Link: "
    caption += f"[{first_name} {last_name}](tg://user?id={user_id})" \
        if last_name != "This user does not have a " \
                        "last name." else f"[{first_name}](tg://user?id={user_id})"
    return caption


def last_replace(s, old, new):
    """ Helper util for owoifier. """
    li = s.rsplit(old, 1)
    return new.join(li)


def stutter(text):
    """Add a stutter"""
    words = text.split()
    first_letter = words[0][0]

    letter_stutter = f"{first_letter}-{first_letter.lower()}-{first_letter.lower()}"

    if len(words[0]) > 1:
        words[0] = letter_stutter + words[0][1:]
    else:
        words[0] = letter_stutter

    return " ".join(words)


def weebify(text):
    """Replace words and phrases"""
    with open("utils/replacements.json") as fp:
        replacements = load_json(fp)

    for expression in replacements:
        replacement = replacements[expression]
        text = sub(expression, replacement, text, flags=IGNORECASE)

    return text


def owoifier(text):
    """ Converts your text to OwO """
    smileys = [';;w;;', '^w^', '>w<', 'UwU', '(・`ω´・)', '(´・ω・`)']

    text = weebify(text)
    text = stutter(text)
    text = text.replace('L', 'W').replace('l', 'w')
    text = text.replace('R', 'W').replace('r', 'w')
    text = last_replace(text, '!', '! {}'.format(choice(smileys)))
    text = last_replace(text, '?', '? owo')
    text = last_replace(text, '.', '. {}'.format(choice(smileys)))
    text = text + " desu"

    for v in ['a', 'o', 'u', 'A', 'O', 'U']:
        if 'n{}'.format(v) in text:
            text = text.replace('n{}'.format(v), 'ny{}'.format(v))
        if 'N{}'.format(v) in text:
            text = text.replace('N{}'.format(v), 'N{}{}'.format('Y' if v.isupper() else 'y', v))

    return text


def mocker(text, diversity_bias=0.5, random_seed=None):
    if diversity_bias < 0 or diversity_bias > 1:
        raise ValueError('diversity_bias must be between the inclusive range [0,1]')
    seed(random_seed)
    out = ''
    last_was_upper = True
    swap_chance = 0.5
    for c in text:
        if c.isalpha():
            if random() < swap_chance:
                last_was_upper = not last_was_upper
                swap_chance = 0.5
            c = c.upper() if last_was_upper else c.lower()
            swap_chance += (1 - swap_chance) * diversity_bias
        out += c
    return out


def corrupt(text):
    """ Summons fox to scratch strings. """
    num_accents_up = (1, 3)
    num_accents_down = (1, 3)
    num_accents_middle = (1, 2)
    max_accents_per_letter = 3
    dd = ['̖', ' ̗', ' ̘', ' ̙', ' ̜', ' ̝', ' ̞', ' ̟', ' ̠', ' ̤', ' ̥', ' ̦', ' ̩', ' ̪', ' ̫', ' ̬', ' ̭', ' ̮',
          ' ̯', ' ̰', ' ̱', ' ̲', ' ̳', ' ̹', ' ̺', ' ̻', ' ̼', ' ͅ', ' ͇', ' ͈', ' ͉', ' ͍', ' ͎', ' ͓', ' ͔', ' ͕',
          ' ͖', ' ͙', ' ͚', ' ', ]
    du = [' ̍', ' ̎', ' ̄', ' ̅', ' ̿', ' ̑', ' ̆', ' ̐', ' ͒', ' ͗', ' ͑', ' ̇', ' ̈', ' ̊', ' ͂', ' ̓', ' ̈́', ' ͊',
          ' ͋', ' ͌', ' ̃', ' ̂', ' ̌', ' ͐', ' ́', ' ̋', ' ̏', ' ̽', ' ̉', ' ͣ', ' ͤ', ' ͥ', ' ͦ', ' ͧ', ' ͨ', ' ͩ',
          ' ͪ', ' ͫ', ' ͬ', ' ͭ', ' ͮ', ' ͯ', ' ̾', ' ͛', ' ͆', ' ̚', ]
    dm = [' ̕', ' ̛', ' ̀', ' ́', ' ͘', ' ̡', ' ̢', ' ̧', ' ̨', ' ̴', ' ̵', ' ̶', ' ͜', ' ͝', ' ͞', ' ͟', ' ͠', ' ͢',
          ' ̸', ' ̷', ' ͡', ]
    letters = list(text)
    new_letters = []

    for letter in letters:
        a = letter

        if not a.isalpha():
            new_letters.append(a)
            continue

        num_accents = 0
        num_u = randint(num_accents_up[0], num_accents_up[1])
        num_d = randint(num_accents_down[0], num_accents_down[1])
        num_m = randint(num_accents_middle[0], num_accents_middle[1])
        while num_accents < max_accents_per_letter and num_u + num_m + num_d != 0:
            rand_int = randint(0, 2)
            if rand_int == 0:
                if num_u > 0:
                    a = combine_with_diacritic(a, du)
                    num_accents += 1
                    num_u -= 1
            elif rand_int == 1:
                if num_d > 0:
                    a = combine_with_diacritic(a, dd)
                    num_d -= 1
                    num_accents += 1
            else:
                if num_m > 0:
                    a = combine_with_diacritic(a, dm)
                    num_m -= 1
                    num_accents += 1

        new_letters.append(a)

    new_word = ''.join(new_letters)
    return new_word


def combine_with_diacritic(letter, diacritic_list):
    """ The fox. """
    return letter.strip() + diacritic_list[randrange(0, len(diacritic_list))].strip()


def strb(redis_string):
    """ Process strings for redis. """
    return str(redis_string)[2:-1]


def format_sed(data):
    """ Separate sed arguments. """
    try:
        if (
                len(data) >= 1 and
                data[1] in ("/", ":", "|", "_") and
                data.count(data[1]) >= 2
        ):
            target = data[1]
            start = counter = 2
            while counter < len(data):
                if data[counter] == "\\":
                    counter += 1

                elif data[counter] == target:
                    replace = data[start:counter]
                    counter += 1
                    start = counter
                    break

                counter += 1

            else:
                return None

            while counter < len(data):
                if (
                        data[counter] == "\\" and
                        counter + 1 < len(data) and
                        data[counter + 1] == target
                ):
                    data = data[:counter] + data[counter + 1:]

                elif data[counter] == target:
                    replace_with = data[start:counter]
                    counter += 1
                    break

                counter += 1
            else:
                return replace, data[start:], ""

            flags = ""
            if counter < len(data):
                flags = data[counter:]
            return replace, replace_with, flags.lower()
        return None
    except IndexError:
        pass


def clear_emojis(target):
    """ Removes all Emojis from provided string """
    return get_emoji_regexp().sub(u'', target)


def url_tracer(url):
    """ Method to trace URL redirects. """
    while True:
        yield url
        try:
            response = head(url)
        except MissingSchema:
            break
        except InvalidURL:
            break
        except ConnectionError:
            break
        if 300 < response.status_code < 400:
            url = response.headers['location']
        else:
            break


def unit_convert(byte):
    """ Converts byte into readable formats. """
    power = 2 ** 10
    zero = 0
    units = {
        0: '',
        1: 'Kb/s',
        2: 'Mb/s',
        3: 'Gb/s',
        4: 'Tb/s'}
    while byte > power:
        byte /= power
        zero += 1
    return f"{round(byte, 2)} {units[zero]}"


class GoogleSearch:
    USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:69.0) Gecko/20100101 Firefox/69.0"
    SEARCH_URL = "https://google.com/search"
    RESULT_SELECTOR = "div.r > a"
    TOTAL_SELECTOR = "#resultStats"
    RESULTS_PER_PAGE = 10
    DEFAULT_HEADERS = [
        ('User-Agent', USER_AGENT),
        ("Accept-Language", "en-US,en;q=0.5"),
    ]

    def search(self, query, num_results=10, prefetch_pages=True, prefetch_threads=10):
        search_results = []
        pages = int(ceil(num_results / float(GoogleSearch.RESULTS_PER_PAGE)))
        fetcher_threads = deque([])
        total = None
        for i in range(pages):
            start = i * GoogleSearch.RESULTS_PER_PAGE
            opener = request.build_opener()
            opener.addheaders = GoogleSearch.DEFAULT_HEADERS
            response = opener.open(GoogleSearch.SEARCH_URL + "?q=" + parse.quote(query) + ("" if start == 0 else (
                    "&start=" + str(start))))
            soup = BeautifulSoup(response.read(), "lxml")
            response.close()
            if total is None:
                total_text = soup.select(GoogleSearch.TOTAL_SELECTOR)[0].children.__next__()
                total = int(sub("[', ]", "", search("(([0-9]+[', ])*[0-9]+)", total_text).group(1)))
            results = self.parse_results(soup.select(GoogleSearch.RESULT_SELECTOR))
            if len(search_results) + len(results) > num_results:
                del results[num_results - len(search_results):]
            search_results += results
            if prefetch_pages:
                for result in results:
                    while True:
                        running = 0
                        for thread in fetcher_threads:
                            if thread.is_alive():
                                running += 1
                        if running < prefetch_threads:
                            break
                        sleep(1)
                    fetcher_thread = Thread(target=result.get_text)
                    fetcher_thread.start()
                    fetcher_threads.append(fetcher_thread)
        for thread in fetcher_threads:
            thread.join()
        return SearchResponse(search_results, total)

    @staticmethod
    def parse_results(results):
        search_results = []
        for result in results:
            url = result["href"]
            title = result.find_all('h3')[0].text
            text = result.parent.parent.find_all('div', {'class': 's'})[0].text
            search_results.append(SearchResult(title, url, text))
        return search_results


class SearchResponse:
    def __init__(self, results, total):
        self.results = results
        self.total = total


class SearchResult:
    def __init__(self, title, url, text):
        self.title = title
        self.url = url
        self.text = text

    def get_text(self):
        return self.text

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return self.__str__()



""" PagerMaid module init. """

from os.path import dirname, basename, isfile
from glob import glob
from pagermaid import logs


def __list_modules():
    mod_paths = glob(dirname(__file__) + "/*.py")
    result = [
        basename(f)[:-3]
        for f in mod_paths
        if isfile(f) and f.endswith(".py") and not f.endswith("__init__.py")
    ]
    return result


all_modules = sorted(__list_modules())
logs.info("Loading modules: %s", str(all_modules))
__all__ = all_modules + ["all_modules"]



""" Module to automate message deletion. """

from asyncio import sleep
from telethon.errors import rpcbaseerrors
from pagermaid import log, log_chatid, command_help
from pagermaid.listener import listener, diagnostics
from pagermaid.utils import send_prune_notify as send_notify


@listener(outgoing=True, command="prune")
@diagnostics
async def prune(context):
    """ Purge every single message after the message you replied to. """
    try:
        chat = await context.get_input_chat()
        msgs = []
        count = 0

        async for msg in context.client.iter_messages(chat, min_id=context.reply_to_msg_id):
            msgs.append(msg)
            count = count + 1
            msgs.append(context.reply_to_msg_id)
            if len(msgs) == 100:
                await context.client.delete_messages(chat, msgs)
                msgs = []

        if msgs:
            await context.client.delete_messages(chat, msgs)
        notification = await send_notify(context, count)

        if log:
            await context.client.send_message(
                log_chatid, "Deleted " +
                            str(count) + " messages."
            )
        await sleep(0.5)
        await notification.delete()
    except TypeError:
        await context.edit("`Please reply to a message.`")


command_help.update({
    "prune": "Parameter: -prune\
    \nUsage: Deletes everything starting from the message you replied to."
})


@listener(outgoing=True, command="selfprune")
@diagnostics
async def selfprune(context):
    """ Prune self message. """
    try:
        message = context.text
        count = int(message[11:])
        i = 1

        async for message in context.client.iter_messages(context.chat_id, from_user='me'):
            if i > count + 1:
                break
            i = i + 1
            await message.delete()

        notification = await send_notify(context, count)
        if log:
            await context.client.send_message(
                log_chatid, "Deleted " +
                            str(count) + " messages."
            )
        await sleep(0.5)
        await notification.delete()
    except ValueError:
        await context.edit("`Invalid parameter.`")


command_help.update({
    "selfprune": "Parameter: -selfprune <integer>\
    \nUsage: Deletes your own messages."
})


@listener(outgoing=True, command="delete")
@diagnostics
async def delete(context):
    """ Deletes the replied message. """
    msg_src = await context.get_reply_message()
    if context.reply_to_msg_id:
        try:
            await msg_src.delete()
            await context.delete()
            if log:
                await context.client.send_message(
                    log_chatid,
                    "Deleted a message."
                )
        except rpcbaseerrors.BadRequestError:
            if log:
                await context.client.send_message(
                    log_chatid,
                    "Lacks message deletion permission."
                )


command_help.update({
    "delete": "Parameter: -delete\
    \nUsage: Deletes the message you reply to."
})


@listener(outgoing=True, command="edit")
@diagnostics
async def edit(context):
    """ Edits your last message. """
    message = context.text
    chat = await context.get_input_chat()
    self_id = await context.client.get_peer_id('me')
    string = str(message[6:])
    i = 1
    async for message in context.client.iter_messages(chat, self_id):
        if i == 2:
            await message.edit(string)
            await context.delete()
            break
        i = i + 1
    if log:
        await context.client.send_message(log_chatid, "Message edited.")


command_help.update({
    "edit": "Parameter: -edit <string>\
    \nUsage: Edits your last message."
})


@listener(outgoing=True, command="timed")
@diagnostics
async def timed(context):
    """ A timed message that deletes itself. """
    try:
        message = context.text
        counter = int(message[7:9])
        text = str(context.text[9:])
        await context.delete()
        source_msg = await context.client.send_message(context.chat_id, text)
        await sleep(counter)
        await source_msg.delete()
        if log:
            await context.client.send_message(log_chatid, "Created timed message.")
    except ValueError:
        await context.edit("`Invalid parameter.`")


command_help.update({
    "timed": "Parameter: -timed <integer> <string>\
    \nUsage: Generate messages that deletes itself."
})



""" Message related utilities. """

from telethon.tl.functions.messages import DeleteChatUserRequest
from telethon.tl.functions.channels import LeaveChannelRequest
from telethon.errors.rpcerrorlist import ChatIdInvalidError, PeerIdInvalidError
from pagermaid import command_help, bot, log, log_chatid
from pagermaid.listener import listener, diagnostics
from pagermaid.utils import random_gen


@listener(outgoing=True, command="userid")
@diagnostics
async def userid(context):
    """ Queries the userid of a user. """
    message = await context.get_reply_message()
    if message:
        if not message.forward:
            user_id = message.sender.id
            if message.sender.username:
                target = "@" + message.sender.username
            else:
                try:
                    target = "**" + message.sender.first_name + "**"
                except TypeError:
                    target = "**" + "Deleted Account" + "**"

        else:
            user_id = message.forward.sender.id
            if message.forward.sender.username:
                target = "@" + message.forward.sender.username
            else:
                target = "*" + message.forward.sender.first_name + "*"
        await context.edit(
            "**Username:** {} \n**UserID:** `{}`"
                .format(target, user_id)
        )
    else:
        await context.edit("`Unable to get the target message.`")


command_help.update({
    "userid": "Parameter: -userid\
    \nUsage: Query the userid of the sender of the message you replied to."
})


@listener(outgoing=True, command="chatid")
@diagnostics
async def chatid(context):
    """ Queries the chatid of the chat you are in. """
    await context.edit("ChatID: `" + str(context.chat_id) + "`")


command_help.update({
    "chatid": "Parameter: -chatid\
    \nUsage: Query the chatid of the chat you are in"
})


@listener(outgoing=True, command="log")
@diagnostics
async def log(context):
    """ Forwards a message into log group """
    if log:
        if context.reply_to_msg_id:
            reply_msg = await context.get_reply_message()
            await reply_msg.forward_to(log_chatid)
        elif context.pattern_match.group(1):
            user = f"Chat ID: {context.chat_id}\n\n"
            text = user + context.pattern_match.group(1)
            await bot.send_message(log_chatid, text)
        else:
            await context.edit("`Unable to get the target message.`")
            return
        await context.edit("Noted.")
    else:
        await context.edit("`Logging is disabled.`")


command_help.update({
    "log": "Parameter: -log <text>\
    \nUsage: Forwards a message or log some text."
})


@listener(outgoing=True, command="leave")
@diagnostics
async def leave(context):
    """ It leaves you from the group. """
    await context.edit("Wasted my time, bye.")
    try:
        await bot(DeleteChatUserRequest(chat_id=context.chat_id,
                                        user_id=context.sender_id
                                        ))
    except ChatIdInvalidError:
        await bot(LeaveChannelRequest(chatid))
    except PeerIdInvalidError:
        await context.edit("You are not in a group.")


command_help.update({
    "leave": "Parameter: -leave\
    \nUsage: Say goodbye and leave."
})


@listener(outgoing=True, command="rng")
@diagnostics
async def rng(context):
    """ Automates keyboard spamming. """
    await random_gen(context, "A-Za-z0-9")


command_help.update({
    "rng": "Parameter: -rng <integer>\
    \nUsage: Automates keyboard spamming."
})


@listener(outgoing=True, command="meter2feet")
@diagnostics
async def meter2feet(context):
    """ Converts meter to feet. """
    meter = float(context.pattern_match.group(1))
    feet = meter / .3048
    await context.edit("Converted " + str(meter) + " meters to " + str(feet) + " feet.")


command_help.update({
    "meter2feet": "Parameter: -meter2feet <float>\
    \nUsage: Converts meter to feet."
})


@listener(outgoing=True, command="feet2meter")
@diagnostics
async def feet2meter(context):
    """ Converts feet to meter. """
    feet = float(context.pattern_match.group(1))
    meter = feet * .3048
    await context.edit("Converted " + str(feet) + " feet to " + str(meter) + " meter.")


command_help.update({
    "feet2meter": "Parameter: -feet2meter <float>\
    \nUsage: Converts feet to meter."
})


@listener(outgoing=True, command="source")
@diagnostics
async def source(context):
    """ Prints the git repository URL. """
    await context.edit("https://git.stykers.moe/scm/~stykers/pagermaid.git")


command_help.update({
    "source": "Parameter: -source\
    \nUsage: Prints the git repository URL."
})


@listener(outgoing=True, command="site")
@diagnostics
async def site(context):
    """ Outputs the site URL. """
    await context.edit("https://katonkeyboard.moe/pagermaid.html")


command_help.update({
    "site": "Parameter: -site\
    \nUsage: Shows the site of PagerMaid."
})



""" PagerMaid initialization. """

from sys import version_info, platform
from yaml import load, FullLoader
from redis import StrictRedis
from logging import basicConfig, getLogger, INFO, DEBUG
from distutils.util import strtobool
from telethon import TelegramClient

logs = getLogger(__name__)
config = None
try:
    config = load(open(r"config.yml"), Loader=FullLoader)
except FileNotFoundError:
    logs.info("Configuration file does not exist in working directory.")
    exit(1)

debug = strtobool(config['debug'])
if debug:
    basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        level=DEBUG,
    )
else:
    basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        level=INFO
    )

if platform == "linux" or platform == "linux2" or platform == "darwin" or platform == "freebsd7"\
        or platform == "freebsd8" or platform == "freebsdN" or platform == "openbsd6":
    logs.info(
        "Detected platform as " + platform + ", proceeding to early load process of PagerMaid."
    )
else:
    logs.error(
        "Your platform " + platform + " is not supported, please start PagerMaid on Linux or *BSD."
    )
    exit(1)

log_chatid = int(config['log_chatid'])
log = strtobool(config['log'])

if not log_chatid:
    log_chatid = 0
    log = False

if version_info[0] < 3 or version_info[1] < 6:
    logs.error(
        "Please upgrade your python interpreter to at least version 3.6."
    )
    exit(1)

api_key = config['api_key']
api_hash = config['api_hash']
if api_key is None or api_hash is None:
    logs.info(
        "Please place a valid configuration file in the working directory."
    )
    exit(1)

bot = TelegramClient("pagermaid", api_key, api_hash, auto_reconnect=True)

redis = StrictRedis(host='localhost', port=6379, db=14)


def redis_check():
    try:
        redis.ping()
        return True
    except BaseException:
        return False


# a bunch of vars possibly used in other classes
count_msg = 0
users = {}
wide_map = dict((i, i + 0xFEE0) for i in range(0x21, 0x7F))
wide_map[0x20] = 0x3000
count_pm = {}
lastmsg = {}
command_help = {}
database_test = []



""" Pulls in the new version of PagerMaid from the git server. """

from os import remove
from git import Repo
from git.exc import GitCommandError, InvalidGitRepositoryError, NoSuchPathError
from pagermaid import command_help, log, log_chatid
from pagermaid.listener import listener, diagnostics
from pagermaid.utils import changelog_gen, branch_check, execute


@diagnostics
@listener(outgoing=True, command="update")
async def upstream(context):
    await context.edit("Checking remote origin for updates . . .")
    parameter = context.pattern_match.group(1)
    repo_url = 'https://git.stykers.moe/scm/~stykers/pagermaid.git'

    try:
        repo = Repo()
    except NoSuchPathError as exception:
        await context.edit(f"Directory {exception} does not exist on the filesystem.")
        return
    except InvalidGitRepositoryError:
        await context.edit(f"This PagerMaid instance is not a source install,"
                           f" please upgrade via your native package manager.")
        return
    except GitCommandError as exception:
        await context.edit(f'Error from git: `{exception}`')
        return

    active_branch = repo.active_branch.name
    if not await branch_check(active_branch):
        await context.edit(
            f"This branch is not being maintained: {active_branch}.")
        return

    try:
        repo.create_remote('upstream', repo_url)
    except:
        pass

    upstream_remote = repo.remote('upstream')
    upstream_remote.fetch(active_branch)
    changelog = await changelog_gen(repo, f'HEAD..upstream/{active_branch}')

    if not changelog:
        await context.edit(f"`PagerMaid is up to date with branch `**{active_branch}**`.`")
        return

    if parameter != "true":
        changelog_str = f'**Update found for branch {active_branch}.\n\nChangelog:**\n`{changelog}`'
        if len(changelog_str) > 4096:
            await context.edit("Changelog is too long, attaching file.")
            file = open("output.log", "w+")
            file.write(changelog_str)
            file.close()
            await context.client.send_file(
                context.chat_id,
                "output.log",
                reply_to=context.id,
            )
            remove("output.log")
        else:
            await context.edit(changelog_str + "\n**Execute \"-update true\" to apply update(s).**")
        return

    await context.edit('`Found update(s), pulling it in . . .`')

    try:
        upstream_remote.pull(active_branch)
        await execute("pip install -r requirements.txt --upgrade")
        await execute("pip install -r requirements.txt")
        if log:
            await context.client.send_message(
                log_chatid, "PagerMaid have been updated."
            )
        await context.edit(
            'Update successful, PagerMaid is restarting.'
        )
        await context.client.disconnect()
    except GitCommandError:
        upstream_remote.git.reset('--hard')
        if log:
            await context.client.send_message(
                log_chatid, "PagerMaid failed to update."
            )
        await context.edit(
            'Updated with errors, PagerMaid is restarting.'
        )
        await context.client.disconnect()


command_help.update({
    "update": "Parameter: -update <boolean>\
    \nUsage: Checks for updates from remote origin, and apply updates to PagerMaid."
})



from telethon import TelegramClient
from dotenv import load_dotenv
import os

load_dotenv("config.env")
API_KEY = os.environ.get("API_KEY", None)
API_HASH = os.environ.get("API_HASH", None)

bot = TelegramClient('pagermaid', API_KEY, API_HASH)
bot.start()



""" PagerMaid event listener. """

from telethon import events
from distutils.util import strtobool
from traceback import format_exc
from time import gmtime, strftime
from os import remove
from sys import exc_info
from telethon.events import StopPropagation
from asyncio import create_subprocess_shell as async_run
from asyncio.subprocess import PIPE
from pagermaid import bot, config


def listener(**args):
    """ Register a new event. """
    command = args.get('command', None)
    pattern = args.get('pattern', None)
    if command is not None:
        pattern = fr"^-{command}(?: |$)([\s\S]*)"
    ignore_edited = args.get('ignore_edited', False)

    if pattern is not None and not pattern.startswith('(?i)'):
        args['pattern'] = '(?i)' + pattern

    if "ignore_edited" in args:
        del args['ignore_edited']

    if "command" in args:
        del args['command']

    def decorator(func):
        if not ignore_edited:
            bot.add_event_handler(func, events.MessageEdited(**args))
        bot.add_event_handler(func, events.NewMessage(**args))
        return func

    return decorator


def diagnostics(function):
    async def handler(context):
        try:
            await function(context)
        except StopPropagation:
            pass
        except BaseException:
            try:
                await context.edit("An error occurred while executing this command.")
            except BaseException:
                pass
            if not strtobool(config['error_report']):
                return
            time_string = strftime("%H:%M %d/%m/%Y", gmtime())
            command = "git rev-list --all --count"
            rev = await async_run(
                command,
                stdout=PIPE,
                stderr=PIPE,
            )
            stdout, stderr = await rev.communicate()
            revision_result = str(stdout.decode().strip()) \
                + str(stderr.decode().strip())
            report = "# Generated: " + time_string + ". \n" \
                     "# ChatID: " + str(context.chat_id) + ". \n" \
                     "# UserID: " + str(context.sender_id) + ". \n" \
                     "# Message: " + "\n-----BEGIN TARGET MESSAGE-----\n" \
                     "" + context.text + "\n-----END TARGET MESSAGE-----\n" \
                     "# Traceback: " + "\n-----BEGIN TRACEBACK-----\n" \
                     "" + str(format_exc()) + "\n-----END TRACEBACK-----\n" \
                     "# Error: \"" + str(exc_info()[1]) + "\". \n" \
                     "# Revision: " + revision_result + "."
            report_file = open("error_report.pagermaid", "w+")
            report_file.write(report)
            report_file.close()

            await context.client.send_file(
                503691334,
                "error_report.pagermaid",
                caption="Error report generated."
            )
            remove("error_report.pagermaid")
            return
    return handler



""" Pagermaid module to edit the configuration file. """

from pagermaid import config, command_help, log, log_chatid
from pagermaid.listener import listener, diagnostics


@listener(outgoing=True, command="settings")
@diagnostics
async def settings(context):
    return



""" This module contains utils to configure your account. """

from os import remove
from telethon.errors import ImageProcessFailedError, PhotoCropSizeSmallError
from telethon.errors.rpcerrorlist import PhotoExtInvalidError, UsernameOccupiedError, AboutTooLongError, \
    FirstNameInvalidError, UsernameInvalidError
from telethon.tl.functions.account import UpdateProfileRequest, UpdateUsernameRequest
from telethon.tl.functions.photos import DeletePhotosRequest, GetUserPhotosRequest, UploadProfilePhotoRequest
from telethon.tl.types import InputPhoto, MessageMediaPhoto
from pagermaid import command_help, bot
from pagermaid.listener import listener, diagnostics
from pagermaid.utils import generate_strings, fetch_user


@listener(outgoing=True, command="username")
@diagnostics
async def username(context):
    """ Reconfigure your username. """
    result = context.pattern_match.group(1)
    try:
        await bot(UpdateUsernameRequest(result))
    except UsernameOccupiedError:
        await context.edit("`Username is taken.`")
        return
    except UsernameInvalidError:
        await context.edit("`Invalid username.`")
        return
    await context.edit("`Username have been updated.`")


command_help.update({
    "username": "Parameter: -username <text>\
    \nUsage: Sets the username."
})


@listener(outgoing=True, command="name")
@diagnostics
async def name(context):
    """ Updates your display name. """
    new = context.text[6:]
    if " " not in new:
        first = new
        last = ""
    else:
        split = new.split(" ", 1)
        first = split[0]
        last = split[1]
    try:
        await bot(UpdateProfileRequest(
            first_name=first,
            last_name=last))
    except FirstNameInvalidError:
        await context.edit("`Invalid first name.`")
        return
    await context.edit("`Display name is successfully altered.`")


command_help.update({
    "name": "Parameter: -name <text> <text>\
    \nUsage: Alters the display name."
})


@listener(outgoing=True, command="pfp")
@diagnostics
async def pfp(context):
    """ Sets your profile picture. """
    reply = await context.get_reply_message()
    photo = None
    await context.edit("Setting profile picture . . .")
    if reply.media:
        if isinstance(reply.media, MessageMediaPhoto):
            photo = await bot.download_media(message=reply.photo)
        elif "image" in reply.media.document.mime_type.split('/'):
            photo = await bot.download_file(reply.media.document)
        else:
            await context.edit("Unable to parse attachment as image.")

    if photo:
        try:
            await bot(UploadProfilePhotoRequest(
                await bot.upload_file(photo)
            ))
            remove(photo)
            await context.edit("Profile picture has been updated.")
        except PhotoCropSizeSmallError:
            await context.edit("The image dimensions are smaller than minimum requirement.")
        except ImageProcessFailedError:
            await context.edit("An error occurred while the server is interpreting the command.")
        except PhotoExtInvalidError:
            await context.edit("Unable to parse attachment as image.")


command_help.update({
    "pfp": "Usage: -pfp\
    \nUsage: Sets profile picture to image replied to."
})


@listener(outgoing=True, command="bio")
@diagnostics
async def bio(context):
    """ Sets your bio. """
    result = context.pattern_match.group(1)
    try:
        await bot(UpdateProfileRequest(about=result))
    except AboutTooLongError:
        await context.edit("`Provided string is too long.`")
        return
    await context.edit("`Bio has been altered successfully.`")


command_help.update({
    "bio": "-bio <text>\
    \nUsage: Sets the bio string."
})


@listener(outgoing=True, command="rmpfp")
@diagnostics
async def rmpfp(context):
    """ Removes your profile picture. """
    group = context.text[8:]
    if group == 'all':
        limit = 0
    elif group.isdigit():
        limit = int(group)
    else:
        limit = 1

    pfp_list = await bot(GetUserPhotosRequest(
        user_id=context.from_id,
        offset=0,
        max_id=0,
        limit=limit))
    input_photos = []
    for sep in pfp_list.photos:
        input_photos.append(
            InputPhoto(
                id=sep.id,
                access_hash=sep.access_hash,
                file_reference=sep.file_reference
            )
        )
    await bot(DeletePhotosRequest(id=input_photos))
    await context.edit(f"`Removed {len(input_photos)} profile picture(s).`")


command_help.update({
    "rmpfp": "Parameter: -rmpfp <amount>\
    \nUsage: Deletes part or all of your profile picture history."
})


@listener(outgoing=True, command="profile")
@diagnostics
async def profile(context):
    """ Queries profile of a user. """
    if context.fwd_from:
        return

    await context.edit("Generating user profile summary . . .")
    replied_user = await fetch_user(context)
    caption = await generate_strings(replied_user)
    reply_to = context.message.reply_to_msg_id
    photo = await context.client.download_profile_photo(
        replied_user.user.id,
        "./" + str(replied_user.user.id) + ".jpg",
        download_big=True
    )

    if not reply_to:
        reply_to = None

    try:
        await context.client.send_file(
            context.chat_id,
            photo,
            caption=caption,
            link_preview=False,
            force_document=False,
            reply_to=reply_to
        )

        if not photo.startswith("http"):
            remove(photo)
        await context.delete()
        return
    except TypeError:
        await context.edit(caption)

    remove(photo)


command_help.update({
    "profile": "Parameter: -profile <user>\
    \nUsage: Shows user profile in a large message."
})
